@@ -15,7 +15,7 @@
  */
 #pragma once
 
-#include "__execution_fwd.hpp" // IWYU pragma: keep
+#include "__execution_fwd.hpp"
 
 #include "__concepts.hpp"
 #include "__cpo.hpp"
@@ -24,8 +24,10 @@
 #include "__tag_invoke.hpp"
 #include "__tuple.hpp"
 
-#include <type_traits>
+#include <exception>  // IWYU pragma: keep for std::terminate
 #include <functional> // IWYU pragma: keep for unwrap_reference_t
+#include <type_traits>
+#include <utility>
 
 STDEXEC_PRAGMA_PUSH()
 STDEXEC_PRAGMA_IGNORE_EDG(probable_guiding_friend)
@@ -37,7 +39,7 @@ namespace stdexec {
   concept queryable = destructible<T>;
 
   template <class Tag>
-  struct __query {
+  struct __query { // NOLINT(bugprone-crtp-constructor-accessibility)
     template <class Sig>
     static inline constexpr Tag (*signature)(Sig) = nullptr;
   };
@@ -46,10 +48,7 @@ namespace stdexec {
   // [exec.queries]
   namespace __queries {
     template <class _Tp>
-    concept __is_bool_constant = //
-      requires {                 //
-        typename __mbool<_Tp::value>;
-      };
+    concept __is_bool_constant = requires { typename __mbool<_Tp::value>; };
 
     struct forwarding_query_t {
       template <class _Query>
@@ -90,7 +89,8 @@ namespace stdexec {
       template <class _Tp>
         requires tag_invocable<execute_may_block_caller_t, __cref_t<_Tp>>
       constexpr auto operator()(_Tp&& __t) const noexcept -> bool {
-        static_assert(same_as<bool, tag_invoke_result_t<execute_may_block_caller_t, __cref_t<_Tp>>>);
+        static_assert(
+          same_as<bool, tag_invoke_result_t<execute_may_block_caller_t, __cref_t<_Tp>>>);
         static_assert(nothrow_tag_invocable<execute_may_block_caller_t, __cref_t<_Tp>>);
         return tag_invoke(execute_may_block_caller_t{}, std::as_const(__t));
       }
@@ -114,24 +114,6 @@ namespace stdexec {
       }
     };
 
-    struct __has_algorithm_customizations_t : __query<__has_algorithm_customizations_t> {
-      template <class _Tp>
-      using __result_t = tag_invoke_result_t<__has_algorithm_customizations_t, __cref_t<_Tp>>;
-
-      template <class _Tp>
-        requires tag_invocable<__has_algorithm_customizations_t, __cref_t<_Tp>>
-      constexpr auto
-        operator()(_Tp&&) const noexcept(noexcept(__result_t<_Tp>{})) -> __result_t<_Tp> {
-        using _Boolean = tag_invoke_result_t<__has_algorithm_customizations_t, __cref_t<_Tp>>;
-        static_assert(_Boolean{} ? true : false); // must be contextually convertible to bool
-        return _Boolean{};
-      }
-
-      constexpr auto operator()(auto&&) const noexcept -> std::false_type {
-        return {};
-      }
-    };
-
     // TODO: implement allocator concept
     template <class _T0>
     concept __allocator_c = true;
@@ -211,8 +193,7 @@ namespace stdexec {
 
     template <class _Queryable, class _Tag>
     concept __has_completion_scheduler_for =
-      queryable<_Queryable> && //
-      tag_invocable<get_completion_scheduler_t<_Tag>, const _Queryable&>;
+      queryable<_Queryable> && tag_invocable<get_completion_scheduler_t<_Tag>, const _Queryable&>;
 
     template <__completion_tag _Tag>
     struct get_completion_scheduler_t : __query<get_completion_scheduler_t<_Tag>> {
@@ -228,7 +209,7 @@ namespace stdexec {
     struct get_domain_t {
       template <class _Ty>
         requires tag_invocable<get_domain_t, const _Ty&>
-      constexpr auto operator()(const _Ty& __ty) const noexcept
+      constexpr auto operator()(const _Ty&) const noexcept
         -> __decay_t<tag_invoke_result_t<get_domain_t, const _Ty&>> {
         static_assert(
           nothrow_tag_invocable<get_domain_t, const _Ty&>,
@@ -244,15 +225,34 @@ namespace stdexec {
       }
     };
 
+    struct get_domain_override_t {
+      template <class _Ty>
+        requires tag_invocable<get_domain_override_t, const _Ty&>
+      constexpr auto operator()(const _Ty&) const noexcept
+        -> __decay_t<tag_invoke_result_t<get_domain_override_t, const _Ty&>> {
+        static_assert(
+          nothrow_tag_invocable<get_domain_override_t, const _Ty&>,
+          "Customizations of get_domain_override must be noexcept.");
+        static_assert(
+          __class<__decay_t<tag_invoke_result_t<get_domain_override_t, const _Ty&>>>,
+          "Customizations of get_domain_override must return a class type.");
+        return {};
+      }
+
+      static constexpr auto query(forwarding_query_t) noexcept -> bool {
+        return false;
+      }
+    };
+
     struct __is_scheduler_affine_t {
       template <class _Env>
-      constexpr auto operator()(const _Env&) const noexcept {
-        if constexpr (tag_invocable<__is_scheduler_affine_t, const _Env&>) {
-          using _Result = tag_invoke_result_t<__is_scheduler_affine_t, const _Env&>;
-          static_assert(__same_as<decltype(__v<_Result>), const bool>);
-          return _Result();
+      constexpr auto operator()(const _Env& __env) const noexcept {
+        if constexpr (requires { _Env::query(*this); }) {
+          return _Env::query(*this);
+        } else if constexpr (requires { __env.query(*this); }) {
+          return __env.query(*this);
         } else {
-          return std::false_type();
+          return false;
         }
       }
 
@@ -287,7 +287,6 @@ namespace stdexec {
   using __queries::forwarding_query_t;
   using __queries::query_or_t;
   using __queries::execute_may_block_caller_t;
-  using __queries::__has_algorithm_customizations_t;
   using __queries::get_forward_progress_guarantee_t;
   using __queries::get_allocator_t;
   using __queries::get_scheduler_t;
@@ -298,6 +297,7 @@ namespace stdexec {
   using __queries::get_stop_token_t;
   using __queries::get_completion_scheduler_t;
   using __queries::get_domain_t;
+  using __queries::get_domain_override_t;
   using __queries::__is_scheduler_affine_t;
   using __queries::__root_t;
   using __queries::__root_env;
@@ -305,7 +305,6 @@ namespace stdexec {
   inline constexpr forwarding_query_t forwarding_query{};
   inline constexpr query_or_t query_or{}; // NOT TO SPEC
   inline constexpr execute_may_block_caller_t execute_may_block_caller{};
-  inline constexpr __has_algorithm_customizations_t __has_algorithm_customizations{};
   inline constexpr get_forward_progress_guarantee_t get_forward_progress_guarantee{};
   inline constexpr get_scheduler_t get_scheduler{};
   inline constexpr get_delegation_scheduler_t get_delegation_scheduler{};
@@ -331,44 +330,73 @@ namespace stdexec {
   concept __forwarding_query = forwarding_query(_Tag{});
 
   inline constexpr get_domain_t get_domain{};
-
-  template <class _Env>
-  using __domain_of_t = __decay_t<__call_result_t<get_domain_t, _Env>>;
+  inline constexpr get_domain_override_t get_domain_override{};
 
   template <class _Tag, class _Queryable, class _Default>
   using __query_result_or_t = __call_result_t<query_or_t, _Tag, _Queryable, _Default>;
 
   namespace __env {
-    // To be kept in sync with the promise type used in __connect_awaitable
-    template <class _Env>
-    struct __promise {
+    template <class _Tp, class _Promise>
+    concept __has_as_awaitable_member = requires(_Tp&& __t, _Promise& __promise) {
+      static_cast<_Tp &&>(__t).as_awaitable(__promise);
+    };
+
+    template <class _Promise>
+    struct __with_await_transform {
       template <class _Ty>
       auto await_transform(_Ty&& __value) noexcept -> _Ty&& {
         return static_cast<_Ty&&>(__value);
       }
 
       template <class _Ty>
-        requires tag_invocable<as_awaitable_t, _Ty, __promise&>
-      auto await_transform(_Ty&& __value) //
-        noexcept(nothrow_tag_invocable<as_awaitable_t, _Ty, __promise&>)
-          -> tag_invoke_result_t<as_awaitable_t, _Ty, __promise&> {
-        return tag_invoke(as_awaitable, static_cast<_Ty&&>(__value), *this);
+        requires __has_as_awaitable_member<_Ty, _Promise&>
+      auto await_transform(_Ty&& __value)
+        noexcept(noexcept(__declval<_Ty>().as_awaitable(__declval<_Promise&>())))
+          -> decltype(__declval<_Ty>().as_awaitable(__declval<_Promise&>())) {
+        return static_cast<_Ty&&>(__value).as_awaitable(static_cast<_Promise&>(*this));
       }
 
+      template <class _Ty>
+        requires(!__has_as_awaitable_member<_Ty, _Promise&>)
+             && tag_invocable<as_awaitable_t, _Ty, _Promise&>
+      auto await_transform(_Ty&& __value)
+        noexcept(nothrow_tag_invocable<as_awaitable_t, _Ty, _Promise&>)
+          -> tag_invoke_result_t<as_awaitable_t, _Ty, _Promise&> {
+        return tag_invoke(as_awaitable, static_cast<_Ty&&>(__value), static_cast<_Promise&>(*this));
+      }
+    };
+
+    template <class _Env>
+    struct __promise : __with_await_transform<__promise<_Env>> {
       auto get_env() const noexcept -> const _Env&;
     };
 
     template <class _Env, class _Query, class... _Args>
-    concept __queryable = //
-      tag_invocable<_Query, const _Env&, _Args...>;
+    concept __queryable = tag_invocable<_Query, const _Env&, _Args...>;
 
     template <class _Env, class _Query, class... _Args>
-    concept __nothrow_queryable = //
-      nothrow_tag_invocable<_Query, const _Env&, _Args...>;
+    concept __nothrow_queryable = nothrow_tag_invocable<_Query, const _Env&, _Args...>;
 
     template <class _Env, class _Query, class... _Args>
-    using __query_result_t = //
-      tag_invoke_result_t<_Query, const _Env&, _Args...>;
+    concept __statically_queryable_i = requires(_Query __q, _Args&&... __args) {
+      std::remove_reference_t<_Env>::query(__q, static_cast<_Args &&>(__args)...);
+    };
+
+    template <class _Env, class _Query, class... _Args>
+    concept __statically_queryable = __queryable<_Env, _Query, _Args...>
+                                  && __statically_queryable_i<_Env, _Query, _Args...>;
+
+    template <class _Env, class _Query, class... _Args>
+    using __query_result_t = tag_invoke_result_t<_Query, const _Env&, _Args...>;
+
+    template <class ValueType>
+    struct __prop_like {
+      template <class _Query>
+      STDEXEC_ATTRIBUTE(nodiscard)
+      constexpr auto query(_Query) const noexcept -> const ValueType& {
+        STDEXEC_TERMINATE();
+      }
+    };
 
     // A singleton environment from a query/value pair
     template <class _Query, class _Value>
@@ -376,45 +404,85 @@ namespace stdexec {
       using __t = prop;
       using __id = prop;
 
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Query __query;
+      static_assert(__callable<_Query, __prop_like<_Value>>);
 
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Value __value;
+      STDEXEC_ATTRIBUTE(no_unique_address) _Query __query;
 
-      STDEXEC_ATTRIBUTE((nodiscard)) constexpr const _Value& query(_Query) const noexcept {
+      STDEXEC_ATTRIBUTE(no_unique_address) _Value __value;
+
+      STDEXEC_ATTRIBUTE(nodiscard) constexpr auto query(_Query) const noexcept -> const _Value& {
         return __value;
       }
-
-      prop& operator=(const prop&) = delete;
     };
 
     template <class _Query, class _Value>
-    prop(_Query, _Value) -> prop<_Query, std::unwrap_reference_t<_Value>>;
+    STDEXEC_HOST_DEVICE_DEDUCTION_GUIDE
+      prop(_Query, _Value) -> prop<_Query, std::unwrap_reference_t<_Value>>;
+
+    template <class _Query, auto _Value>
+    struct cprop {
+      using __t = cprop;
+      using __id = cprop;
+
+      STDEXEC_ATTRIBUTE(nodiscard)
+      static constexpr auto query(_Query) noexcept {
+        return _Value;
+      }
+    };
 
     // utility for joining multiple environments
+    template <class... _Envs>
+    struct env;
+
+    template <>
+    struct env<> {
+      using __t = env;
+      using __id = env;
+    };
+
     template <class... _Envs>
     struct env {
       using __t = env;
       using __id = env;
 
-      __tuple_for<_Envs...> __tup_;
+      __tuple_for<_Envs..., env<>> __tup_;
 
       // return a reference to the first child env for which
       // __queryable<_Envs, _Query, _Args...> is true.
       template <class _Query, class... _Args>
-      STDEXEC_ATTRIBUTE((always_inline)) constexpr decltype(auto) __get_1st() const noexcept {
-        constexpr bool __flags[] = {__queryable<_Envs, _Query, _Args...>...};
+      STDEXEC_ATTRIBUTE(always_inline)
+      static constexpr auto __get_1st(const env& __self) noexcept -> decltype(auto) {
+        // NOLINTNEXTLINE (modernize-avoid-c-arrays)
+        constexpr bool __flags[] = {__queryable<_Envs, _Query, _Args...>..., true};
         constexpr std::size_t __idx = __pos_of(__flags, __flags + sizeof...(_Envs));
-        return __tup::get<__idx>(__tup_);
+        return __self.__tup_.template __get<__idx>(__self.__tup_);
       }
 
       template <class _Query, class... _Args>
-        requires(__queryable<_Envs, _Query, _Args...> || ...)
-      STDEXEC_ATTRIBUTE((always_inline)) constexpr decltype(auto) query(_Query __q, _Args&&... __args) const
-        noexcept(__nothrow_queryable<decltype(__get_1st<_Query, _Args...>()), _Query, _Args...>) {
-        return tag_invoke(__q, __get_1st<_Query, _Args...>(), static_cast<_Args&&>(__args)...);
+      using __1st_env_t = decltype(env::__get_1st<_Query, _Args...>(__declval<const env&>()));
+
+      // NOT TO SPEC: a static query memfn for those envs that have a static query memfn.
+      // This is useful for constexpr evaluation of queries.
+      template <class _Query, class... _Args>
+        requires __statically_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>
+      STDEXEC_ATTRIBUTE(nodiscard, always_inline)
+      static constexpr auto query(_Query __q, _Args&&... __args)
+        noexcept(__nothrow_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+          -> decltype(auto) {
+        return std::remove_reference_t<__1st_env_t<_Query, _Args...>>::query(
+          __q, static_cast<_Args&&>(__args)...);
       }
 
-      env& operator=(const env&) = delete;
+      template <class _Query, class... _Args>
+        requires __queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>
+              && (!__statically_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+      STDEXEC_ATTRIBUTE(nodiscard, always_inline)
+      constexpr auto query(_Query __q, _Args&&... __args) const
+        noexcept(__nothrow_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+          -> decltype(auto) {
+        return tag_invoke(
+          __q, env::__get_1st<_Query, _Args...>(*this), static_cast<_Args&&>(__args)...);
+      }
     };
 
     // specialization for two envs to avoid warnings about elided braces
@@ -423,38 +491,56 @@ namespace stdexec {
       using __t = env;
       using __id = env;
 
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Env0 __env0_;
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Env1 __env1_;
+      STDEXEC_ATTRIBUTE(no_unique_address) _Env0 __env0_;
+      STDEXEC_ATTRIBUTE(no_unique_address) _Env1 __env1_;
 
       // return a reference to the first child env for which
       // __queryable<_Envs, _Query, _Args...> is true.
       template <class _Query, class... _Args>
-      STDEXEC_ATTRIBUTE((always_inline)) constexpr decltype(auto) __get_1st() const noexcept {
+      STDEXEC_ATTRIBUTE(always_inline)
+      static constexpr auto __get_1st(const env& __self) noexcept -> decltype(auto) {
         if constexpr (__queryable<_Env0, _Query, _Args...>) {
-          return (__env0_);
+          return (__self.__env0_);
         } else {
-          return (__env1_);
+          return (__self.__env1_);
         }
       }
 
       template <class _Query, class... _Args>
-        requires __queryable<_Env0, _Query, _Args...> || __queryable<_Env1, _Query, _Args...>
-      STDEXEC_ATTRIBUTE((always_inline)) constexpr decltype(auto) query(_Query __q, _Args&&... __args) const
-        noexcept(__nothrow_queryable<decltype(__get_1st<_Query, _Args...>()), _Query, _Args...>) {
-        return tag_invoke(__q, __get_1st<_Query, _Args...>(), static_cast<_Args&&>(__args)...);
+      using __1st_env_t = decltype(env::__get_1st<_Query, _Args...>(__declval<const env&>()));
+
+      // NOT TO SPEC: a static query memfn for those envs that have a static query memfn.
+      // This is useful for constexpr evaluation of queries.
+      template <class _Query, class... _Args>
+        requires __statically_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>
+      STDEXEC_ATTRIBUTE(nodiscard, always_inline)
+      static constexpr auto query(_Query __q, _Args&&... __args)
+        noexcept(__nothrow_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+          -> decltype(auto) {
+        return std::remove_reference_t<__1st_env_t<_Query, _Args...>>::query(
+          __q, static_cast<_Args&&>(__args)...);
       }
 
-      env& operator=(const env&) = delete;
+      template <class _Query, class... _Args>
+        requires __queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>
+              && (!__statically_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+      STDEXEC_ATTRIBUTE(nodiscard, always_inline)
+      constexpr auto query(_Query __q, _Args&&... __args) const
+        noexcept(__nothrow_queryable<__1st_env_t<_Query, _Args...>, _Query, _Args...>)
+          -> decltype(auto) {
+        return tag_invoke(
+          __q, env::__get_1st<_Query, _Args...>(*this), static_cast<_Args&&>(__args)...);
+      }
     };
 
     template <class... _Envs>
-    env(_Envs...) -> env<std::unwrap_reference_t<_Envs>...>;
+    STDEXEC_HOST_DEVICE_DEDUCTION_GUIDE env(_Envs...) -> env<std::unwrap_reference_t<_Envs>...>;
 
     template <class _Value, class _Tag, class... _Tags>
     struct __with {
       using __t = __with;
       using __id = __with;
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Value __value_;
+      STDEXEC_ATTRIBUTE(no_unique_address) _Value __value_;
 
       __with() = default;
 
@@ -462,8 +548,8 @@ namespace stdexec {
         : __value_(static_cast<_Value&&>(__value)) {
       }
 
-      constexpr explicit __with(_Value __value, _Tag, _Tags...) noexcept(
-        __nothrow_decay_copyable<_Value>)
+      constexpr explicit __with(_Value __value, _Tag, _Tags...)
+        noexcept(__nothrow_decay_copyable<_Value>)
         : __value_(static_cast<_Value&&>(__value)) {
       }
 
@@ -471,23 +557,26 @@ namespace stdexec {
       auto query(_Key) const noexcept -> const _Value& {
         return __value_;
       }
-
-      __with& operator=(const __with&) = delete;
     };
 
     template <class _Value, class _Tag, class... _Tags>
     __with(_Value, _Tag, _Tags...) -> __with<_Value, _Tag, _Tags...>;
 
+    template <class _Env>
+    struct __fwd_base {
+      using __fwd_env_t = _Env;
+    };
+
     template <class _EnvId>
     struct __fwd {
       using _Env = __cvref_t<_EnvId>;
       static_assert(__nothrow_move_constructible<_Env>);
 
-      struct __t {
+      struct __t : __fwd_base<_Env> {
         using __id = __fwd;
-        STDEXEC_ATTRIBUTE((no_unique_address)) _Env __env_;
+        STDEXEC_ATTRIBUTE(no_unique_address) _Env __env_;
 
-#if STDEXEC_GCC() && __GNUC__ < 12
+#if STDEXEC_GCC() && STDEXEC_GCC_VERSION < 12'00
         using __cvref_env_t = std::add_const_t<_Env>&;
 #else
         using __cvref_env_t = const _Env&;
@@ -499,22 +588,30 @@ namespace stdexec {
           -> tag_invoke_result_t<_Tag, __cvref_env_t> {
           return tag_invoke(_Tag(), __env_);
         }
-
-        __t& operator=(const __t&) = delete;
       };
     };
 
+    template <class _Env>
+    concept __is_fwd_env = same_as<_Env, typename _Env::__fwd_env_t>;
+
     struct __fwd_fn {
       template <class _Env>
-      auto operator()(_Env&& __env) const {
-        return __t<__fwd<__cvref_id<_Env>>>{static_cast<_Env&&>(__env)};
+      auto operator()(_Env&& __env) const -> decltype(auto) {
+        if constexpr (__is_fwd_env<__decay_t<_Env>>) {
+          return static_cast<_Env>(static_cast<_Env&&>(__env));
+        } else {
+          return __t<__fwd<__cvref_id<_Env>>>{{}, static_cast<_Env&&>(__env)};
+        }
       }
 
-      auto operator()(empty_env) const -> empty_env {
+      auto operator()(env<>) const -> env<> {
         return {};
       }
     };
 
+    template <class _Env>
+    using __fwd_env_t = __call_result_t<__fwd_fn, _Env>;
+
     template <class _EnvId, class _Tag>
     struct __without_ {
       using _Env = __cvref_t<_EnvId>;
@@ -524,7 +621,7 @@ namespace stdexec {
         using __id = __without_;
         _Env __env_;
 
-#if STDEXEC_GCC() && __GNUC__ < 12
+#if STDEXEC_GCC() && STDEXEC_GCC_VERSION < 12'00
         using __cvref_env_t = std::add_const_t<_Env>&;
 #else
         using __cvref_env_t = const _Env&;
@@ -533,12 +630,11 @@ namespace stdexec {
         auto query(_Tag) const noexcept = delete;
 
         template <tag_invocable<__cvref_env_t> _Key>
-        STDEXEC_ATTRIBUTE((always_inline)) auto
-          query(_Key) const noexcept(nothrow_tag_invocable<_Key, __cvref_env_t>) -> decltype(auto) {
+        STDEXEC_ATTRIBUTE(always_inline)
+        auto query(_Key) const noexcept(nothrow_tag_invocable<_Key, __cvref_env_t>)
+          -> decltype(auto) {
           return tag_invoke(_Key(), __env_);
         }
-
-        __t& operator=(const __t&) = delete;
       };
     };
 
@@ -563,7 +659,7 @@ namespace stdexec {
     struct __from {
       using __t = __from;
       using __id = __from;
-      STDEXEC_ATTRIBUTE((no_unique_address)) _Fun __fun_;
+      STDEXEC_ATTRIBUTE(no_unique_address) _Fun __fun_;
 
       template <class _Tag>
         requires __callable<const _Fun&, _Tag>
@@ -571,31 +667,29 @@ namespace stdexec {
         -> __call_result_t<const _Fun&, _Tag> {
         return __fun_(_Tag());
       }
-
-      __from& operator=(const __from&) = delete;
     };
 
     template <class _Fun>
     __from(_Fun) -> __from<_Fun>;
 
     struct __join_fn {
-      auto operator()(empty_env, empty_env) const noexcept -> empty_env {
+      auto operator()(env<>, env<>) const noexcept -> env<> {
         return {};
       }
 
       template <class _Env>
-      auto operator()(_Env&& __env, empty_env = {}) const noexcept -> _Env {
+      auto operator()(_Env&& __env, env<> = {}) const noexcept -> _Env {
         return static_cast<_Env&&>(__env);
       }
 
       template <class _Env>
-      auto operator()(empty_env, _Env&& __env) const noexcept -> decltype(auto) {
+      auto operator()(env<>, _Env&& __env) const noexcept -> decltype(auto) {
         return __fwd_fn()(static_cast<_Env&&>(__env));
       }
 
       template <class _First, class _Second>
       auto operator()(_First&& __first, _Second&& __second) const noexcept
-        -> env<_First, __call_result_t<__fwd_fn, _Second>> {
+        -> env<_First, __fwd_env_t<_Second>> {
         return {static_cast<_First&&>(__first), __fwd_fn()(static_cast<_Second&&>(__second))};
       }
     };
@@ -603,12 +697,12 @@ namespace stdexec {
     inline constexpr __join_fn __join{};
 
     template <class _First, class... _Second>
-    using __join_t = __result_of<__join, _First, _Second...>;
+    using __join_env_t = __result_of<__join, _First, _Second...>;
 
     struct __as_root_env_fn {
       template <class _Env>
       constexpr auto operator()(_Env __env) const noexcept
-        -> __join_t<__root_env, std::unwrap_reference_t<_Env>> {
+        -> __join_env_t<__root_env, std::unwrap_reference_t<_Env>> {
         return __join(__root_env{}, static_cast<std::unwrap_reference_t<_Env>&&>(__env));
       }
     };
@@ -619,24 +713,32 @@ namespace stdexec {
     using __as_root_env_t = __result_of<__as_root_env, _Env>;
   } // namespace __env
 
-  using __env::prop;
-  using __env::env;
-  using empty_env = env<>;
+  using __env::__join_env_t;
+  using __env::__fwd_env_t;
 
   /////////////////////////////////////////////////////////////////////////////
   namespace __get_env {
+    template <class _EnvProvider>
+    concept __has_get_env = requires(const _EnvProvider& __env_provider) {
+      __env_provider.get_env();
+    };
+
     // For getting an execution environment from a receiver or the attributes from a sender.
     struct get_env_t {
-      template <__same_as<get_env_t> _Self, class _EnvProvider>
-      STDEXEC_ATTRIBUTE((always_inline)) friend auto tag_invoke(_Self, const _EnvProvider& __env_provider) noexcept
+      template <class _EnvProvider>
+        requires __has_get_env<_EnvProvider>
+      STDEXEC_ATTRIBUTE(always_inline)
+      constexpr auto operator()(const _EnvProvider& __env_provider) const noexcept
         -> decltype(__env_provider.get_env()) {
+        static_assert(queryable<decltype(__env_provider.get_env())>);
         static_assert(noexcept(__env_provider.get_env()), "get_env() members must be noexcept");
         return __env_provider.get_env();
       }
 
       template <class _EnvProvider>
-        requires tag_invocable<get_env_t, const _EnvProvider&>
-      STDEXEC_ATTRIBUTE((always_inline)) constexpr auto operator()(const _EnvProvider& __env_provider) const noexcept
+        requires(!__has_get_env<_EnvProvider>) && tag_invocable<get_env_t, const _EnvProvider&>
+      STDEXEC_ATTRIBUTE(always_inline)
+      constexpr auto operator()(const _EnvProvider& __env_provider) const noexcept
         -> tag_invoke_result_t<get_env_t, const _EnvProvider&> {
         static_assert(queryable<tag_invoke_result_t<get_env_t, const _EnvProvider&>>);
         static_assert(nothrow_tag_invocable<get_env_t, const _EnvProvider&>);
@@ -644,7 +746,7 @@ namespace stdexec {
       }
 
       template <class _EnvProvider>
-      constexpr auto operator()(const _EnvProvider&) const noexcept -> empty_env {
+      constexpr auto operator()(const _EnvProvider&) const noexcept -> env<> {
         return {};
       }
     };
@@ -654,10 +756,59 @@ namespace stdexec {
   inline constexpr get_env_t get_env{};
 
   template <class _EnvProvider>
-  concept environment_provider = //
-    requires(_EnvProvider& __ep) {
-      { get_env(std::as_const(__ep)) } -> queryable;
-    };
+  concept environment_provider = requires(_EnvProvider& __ep) {
+    { get_env(std::as_const(__ep)) } -> queryable;
+  };
+
+  template <class _Scheduler, class _LateDomain = __none_such>
+  struct __sched_attrs {
+    using __t = __sched_attrs;
+    using __id = __sched_attrs;
+
+    using __scheduler_t = __decay_t<_Scheduler>;
+    using __sched_domain_t = __query_result_or_t<get_domain_t, __scheduler_t, default_domain>;
+    _Scheduler __sched_;
+    STDEXEC_ATTRIBUTE(no_unique_address) _LateDomain __late_domain_ { };
+
+    auto query(get_completion_scheduler_t<set_value_t>) const noexcept -> __scheduler_t {
+      return __sched_;
+    }
+
+    constexpr auto query(get_domain_t) const noexcept -> __sched_domain_t {
+      return {};
+    }
+
+    constexpr auto query(get_domain_override_t) const noexcept -> _LateDomain
+      requires(!same_as<_LateDomain, __none_such>)
+    {
+      return {};
+    }
+  };
+
+  template <class _Scheduler, class _LateDomain = __none_such>
+  __sched_attrs(_Scheduler, _LateDomain = {})
+    -> __sched_attrs<std::unwrap_reference_t<_Scheduler>, _LateDomain>;
+
+  template <class _Scheduler>
+  struct __sched_env {
+    using __t = __sched_env;
+    using __id = __sched_env;
+
+    using __scheduler_t = __decay_t<_Scheduler>;
+    using __sched_domain_t = __query_result_or_t<get_domain_t, __scheduler_t, default_domain>;
+    _Scheduler __sched_;
+
+    auto query(get_scheduler_t) const noexcept -> __scheduler_t {
+      return __sched_;
+    }
+
+    auto query(get_domain_t) const noexcept -> __sched_domain_t {
+      return {};
+    }
+  };
+
+  template <class _Scheduler>
+  __sched_env(_Scheduler) -> __sched_env<std::unwrap_reference_t<_Scheduler>>;
 
   using __env::__as_root_env_t;
   using __env::__as_root_env;
@@ -668,8 +819,9 @@ namespace stdexec {
   };
 
   template <class _Sender>
-  concept __is_scheduler_affine = //
-    requires { requires __v<__call_result_t<__is_scheduler_affine_t, env_of_t<_Sender>>>; };
+  concept __is_scheduler_affine = requires {
+    requires std::remove_reference_t<env_of_t<_Sender>>::query(__is_scheduler_affine_t{});
+  };
 } // namespace stdexec
 
 STDEXEC_PRAGMA_POP()
